
	global _init_omac_1, _omac_1

	section .data

filler  dc.b 80h,0,0,0, 0,0,0,0   ; 100000000... binary

	section .bss

omac_1_state          ; 32 bytes of state
Lu      ds.b 8
Lu2     ds.b 8
key     ds.b 16

	section .text
	align 1

_init_omac_1          ; void init_omac_1(void *key_in_E);

	ldd  # 16         ; copy cipher key
	pshm d
	ldd  # key        ; and key_in_E
	jsr  _bcopy
	ais  #+2

	lde  # Lu2        ; utilizing Lu2 as a tmp zero buffer
	ldd  # 8
	jsr  _bzero       ; first, cipher a block of zeroes

	lde  # Lu2
	ldd  # key
	jsr  _xtea_cipher ; ciphers in-place, E(K,0^n)

	ldx  # Lu2 + 8
	ldy  # Lu  + 8    ; determine Lu
	bsr  LuLu2

	ldx  # Lu  + 8
	ldy  # Lu2 + 8    ; determine Lu2

	; tail. fall thru

LuLu2
	ldaa # 8          ; nbytes
	cmpa # 0          ; clear carry (it is not affected by aix or deca)
L0
	aix  # -1         ; index from 7 to 0
	aiy  # -1
	ldab 0,x
	rolb
	stab 0,y
	deca
	bne  L0           ; until at index 0
	bcc  L1           ; test carry from last rolb
	ldab # 1Bh        ; Constant for n=64
	eorb 7,y
	stab 7,y
L1
	rts


	; rummage thru data chunks in 'v'
	; omac-1 result leaved to 'result'

_omac_1               ; (byte *result_buffer, struct { int n; byte *p; } *v)
	pshm z
	pshm e,d          ; result block, input vector (d goes first)
	ais  #-4          ; _cc and _ptr
	tsz
_cc  equ 0            ; input bytes remaining in current chunk
_ptr equ 2            ; next byte in current chunk
_buf equ 4            ; output/work buffer
_v   equ 6            ; input vector remaining

	ldd  # 8          ; e still has 'result'
	jsr  _bzero       ; work buffer, 8 bytes, zero initially 

	ldx  _v,z         ; first chunk
	ldd  0,x
	std  _cc,z        ; last chunk (not this !) has 0 length
	ldd  2,x
	std  _ptr,z
L2
	ldx  _ptr,z       ; chunk remaining
	bsr  xorbuf       ; merge work buffer and block of input
	stx  _ptr,z

	ldx  # Lu2        ; assumption
	tstw _cc,z        ; cc is signed. no gargantuan chunks, please.
	bgt  L4           ; not last block yet, skip
	blt  L3           ; partial last block, use Lu2
	ldx  # Lu         ; use Lu instead, when cc multiple of 8
L3
	bsr  xorbuf       ; add Lxxx into soup. here, _cc <= 0 already.
L4
	lde  _buf,z       ; work buffer
	ldd  # key
	jsr  _xtea_cipher ; does 8 bytes in-place
	tstw _cc,z
	bgt  L2           ; while there are more bytes

	ais  #+8          ; get rid of locals
	pulm z
	rts

	; xor next 8 bytes of input data or Lu or Lu2 into work buffer.
	; straddle chunks as needed.
	; add padding when end of input.
	; when adding Lxxx, _cc is <= 0 already

xorbuf
	ldy  _buf,z       ; work buffer
	ldaa # 8          ; 8 bytes
L5
	ldab 0,x          ; input
	aix  #+1
	eorb 0,y          ; work
	stab 0,y
	aiy  #+1
	decw _cc,z        ; catch when input chunk ends
	bne  L7

	ldx  _v,z         ; another chunk ?
	aix  # 4
	stx  _v,z         ; ++v
	lde  0,x          ; v->n == 0, if end of vector
	ldx  2,x          ; v->p
	tste
	bne  L6
	ldx  # filler     ; data finis. flip to padding
L6
	ste  _cc,z
L7
	deca
	bne  L5           ; 8 bytes in buffer, padding or not.
	rts

