
#include "param.h"
#include "systm.h"
#include "malloc.h"
#include "stat.h"
#include "ioctl.h"
#include "socket.h"
#include "socketvar.h"
#include "protosw.h"
#include "domain.h"
#include "mbuf.h"

#include "net/if.h"
#include "net/route.h"
#include "net/af.h"
#include "net/netisr.h"

#include "netinet/in.h"
#include "netinet/in_arp.h"
#include "netinet/in_var.h"

#include "ax25/ax.h"
#include "ax25/ax25.h"
#include "ax25/ax_pcb.h"
#include "ax25/ax_var.h"

struct ifqueue axintrq;

axintr()
{
	struct mbuf *m;
	struct ifnet *ifp;
	int s;

	for (;;) {
		s = splimp();
		IF_DEQUEUE(&axintrq, m);
		splx(s);
		if (m == NULL)
			break;

		ax_input(m);
	}
}

ax_input(m)
	struct mbuf *m;
{
#define ac ((struct arpcom *) ifp)
	struct ax_addr *addr, *digis;
	struct ifnet *ifp;
	int s, forus, naddr, len, pull_add;
	u_char ctrl, pid;

	++axstat.axs_total;

	if ((m->m_flags & M_PKTHDR) == 0)
		panic("ax_input no PKTHDR");

	ifp = m->m_pkthdr.rcvif;
	len = m->m_pkthdr.len;

	if (len < AXMINLEN) {
		m_freem(m);
		axstat.axs_toosmall++;
		return;
	}
	if (len == AXMINLEN) {
		pull_add = 1;
	} else {
		pull_add = 2;
	}
	if (m->m_len < 2 * AXALEN + pull_add) {
		m = m_pullup(m, 2 * AXALEN + pull_add);
		if(m == NULL) {
			axstat.axs_nobufs++;
			return;
		}
	}
	if (ax_debug) {
		ax_trace("axi", m);
	}
	{
		extern int ax_monitored;

		if (ax_monitored)
			ax_monitor("axi", m);
	}
	naddr = 1;
	while ((mtod(m, struct ax_addr *)[naddr++].ax_ssid & SSID_E) == 0) {
		if (len < naddr * AXALEN + pull_add) {
			m_freem(m);
			axstat.axs_toosmall++;
			return;
		}
		m = m_pullup(m, naddr * AXALEN + pull_add);
		if (m == NULL) {
			axstat.axs_nobufs++;
			return;
		}
	}
	addr = mtod(m, struct ax_addr *);
	ctrl = ((u_char *) (&addr[naddr]))[0];
	if (pull_add > 1)
		pid = ((u_char *) (&addr[naddr]))[1];
	else
		pid = AXTYPE_NOL3;

	if (naddr > 2)
		digis = &addr[2];
	else
		digis = NULL;

	switch (pid) {

	case AXTYPE_ARP:
		m_adj(m, naddr * AXALEN + 2);
		arpinput(ac, m);
		return;

	case AXTYPE_NOL3:
	case AXTYPE_IP:
		break;

	default:
		axraw_input(ifp, pid, m);
		return;
	}
	forus = 0;
	if (ax_addreq(ac->ac_hwaddr, &addr[0])) {
		forus = 1;
		m->m_flags &= ~(M_BCAST|M_MCAST);
	} else if (ax_ismulticast(&addr[0])) {
		forus = 1;
		m->m_flags |= M_BCAST;
	}
	if (m->m_flags & (M_BCAST|M_MCAST))
		ifp->if_imcasts++;

	switch (pid) {

	case AXTYPE_NOL3:
		if (!forus)
			break;

		m_adj(m, naddr * AXALEN + pull_add);

		if (m->m_flags & (M_BCAST|M_MCAST))
			beacon_input(ifp, &addr[0], &addr[1], digis, ctrl, pid, m);
		else
			lapb_input(ifp, &addr[0], &addr[1], digis, ctrl, pid, m);
		return;

	case AXTYPE_IP:
		if (!forus) {
			m_freem(m);
			return;
		}
		m_adj(m, naddr * AXALEN + 2);
		s = splimp();
		schednetisr(NETISR_IP);
		if (IF_QFULL(&ipintrq)) {
			IF_DROP(&ipintrq);
			m_freem(m);
		} else
			IF_ENQUEUE(&ipintrq, m);
		splx(s);
		return;
	}
	axraw_input(ifp, pid, m);
}

