
/*
 *  Michael Palin: "Is it raining ?"
 */

#include "standard.h"
#include "hamdr_cfg.h"
#include "i2c.h"

#define WX_GLOBAL
#include "wx.h"

#define WX_INTERVAL 10   /* seconds between samples */


#define MINUTES(n) (n * 60 / WX_INTERVAL)

/* number of anemopulses received, x- and y-wise, signed */

STATIC struct windcomp {
	long x, y;
} veclog[MINUTES(1)], *veclogger = veclog;

/* miles/hour in speedlog[] */

STATIC word speedlog[MINUTES(5)], *speedlogger = speedlog;

extern dword
	ow_accvalue(word rank),
	read_pulsacc(stamp *time);

/* XXX handle sunlight effect */

short
	temperature_lowest  = SHORT_MAX,
	temperature_highest = SHORT_MIN;

void
thermometer_hilo_rewind()
{
	temperature_lowest  = SHORT_MAX;
	temperature_highest = SHORT_MIN;
}

void
thermometer(int half_c)
{
	if (half_c < temperature_lowest) {
		temperature_lowest = half_c;
		temperature_celsius_lo    = half_c >> 1;
		temperature_fahrenheit_lo = half_c_to_fahrenheit(half_c);
	}
	if (half_c > temperature_highest) {
		temperature_highest = half_c;
		temperature_celsius_hi    = half_c >> 1;
		temperature_fahrenheit_hi = half_c_to_fahrenheit(half_c);
	}
	have_thermometer = TRUE;
}

void
hygrometer(word rh)
{
	relative_humidity = rh;

	have_hygrometer = TRUE;
}

/* 5 minute peak gust */

word
windgust()
{
	word max = 0, *p = speedlog; /* need not traverse in chronological order */

	do {
		if (max < *p)
			max = *p;
	} while (++p < __ENDOF(speedlog));

	return (max);
}

/* 1 minute sustained, recent ones from speedlog. #define "sustained" XXX */

word
windspeed()
{
	word sum = 0;           /* "sustained" == average (?) */
	word *p = speedlogger;  /* one past last */

	byte n = MINUTES(1);    /* assume less than 256 samples in 1 minute */

	do {
		if (--p < speedlog)
			p = __ENDOF(speedlog) - 1; /* wrap to last */
		sum += *p;
	} while (--n);

	return (sum / MINUTES(1));
}

/* 1 minute average */

word
winddir()
{
	long ew = 0, ns = 0;
	struct windcomp *p = veclog; /* no need for chronological traversal */

	/* cumulate north/south and east/west totals */

	do {
		ew += p->x;
		ns += p->y;
	} while (++p < __ENDOF(veclog));

	return (iatan2(ew, ns));
}

/* relative sine component of wind from direction 'dir', magnitude 'magn' */

static long
jorgio(word deg, word magn)
{
	bool neg;
	long x;

	if (magn == 0)
		magn = 1;    /* avoid loss of information */

	if (neg = (deg >= 180))
		deg -= 180;

	x = EMUL(magn, eight_bit_sin_from_abs_degrees(deg));

	if (neg)
		lnegate(&x);

	return (x);
}

/* 'magn' is "amount of wind received lately" */

static void
log_wind_sample(word deg, word magn, word mph)
{
	long x, y;              /* wind components */

	x = jorgio(deg, magn);  /* sine */

	deg += 90;              /* offset to get the ... */
	if (deg >= 360)
		deg -= 360;

	y = jorgio(deg, magn);  /* ... cosine */

	hold_txproc();

	veclogger->x = x;
	veclogger->y = y;

	if (++veclogger >= __ENDOF(veclog))
		veclogger = veclog;

	*speedlogger = mph;

	if (++speedlogger >= __ENDOF(speedlog))
		speedlogger = speedlog;

	okay_txproc();
}

/*
 *  a common transducer has 10 Hz per 1 m/s wind
 *  our reports must use knots for windspeed and mph for gusts.
 *  using mph internally.
 *  assuming 10 second nominal interval:
 *
 *  50 mph, 22.35 m/s, 2235 ticks / 10 seconds
 *  2231 * 2237 / 10'000 / 10 = 49
 *  2277 * 2237 / 10'000 / 10 = 51
 *  1 mph, 0.4470 m/s, 44 ticks / 10 seconds
 *  44 * 2237 / 10'000 / 10 = 0
 *  45 * 2237 / 10'000 / 10 = 1
 *  sigh. good enough
 */

/*
 *  put new pulses into *count
 *  return elapsed msec (nominally 10'000).
 */
static word
anemometer(word *count)
{
	static stamp mark;
	static dword old;

	stamp now;
	dword n;

	n = read_pulsacc(&now) - old;
	old += n;
	if (n > 65535)
		n = 65535;     /* absurd speed, or time problem */
	*count = n;        /* new pulses */

	n = now - mark;
	mark = now;
	if (n > 65535)
		n = 0;         /* eek? too long since last. skip this sample */
	return (n);        /* elapsed msec */
}

/* de-Gray XXX hw-specific */

static word
vane_to_deg[] = {
	  0,   6,  17,  11,  39,  34,  23,  28,
	 84,  79,  68,  73,  45,  51,  62,  56,
	174, 169, 158, 163, 135, 141, 152, 146,
	 90,  96, 107, 101, 129, 124, 113, 118,
	354, 349, 338, 343, 315, 321, 332, 326,
	270, 276, 287, 281, 309, 304, 293, 298,
	180, 186, 197, 191, 219, 214, 203, 208,
	264, 259, 248, 253, 225, 231, 242, 236
};

/* interval is 10 seconds or less, bytes will do */

void
log_rain_gauge(word pulses)
{
	static byte log[MINUTES(60)];
	static word i;
	word undo;

	rainfall_today += pulses;

	if (pulses > 255)
		pulses = 255;

	undo = log[i];
	log[i] = pulses;
	pulses -= undo;
	rainfall_1hr += pulses;

	if (++i == __SIZEOF(log))
		i = 0;

	have_rain_gauge = TRUE;
}

word
rain_gauge_centi_inch(word pulses)
{
	return (muldiv(pulses, 100, cfg.rainscale));
}

/*
 *  called every WX_INTERVAL seconds
 *
 *  determine current wind speed and direction,
 *  log the values for averaging. maybe print to serial.
 *
 *  The "/ 10" in mph calculation below, is part of anemoscale,
 *  _not_ related to the (10 second ?) interval.
 *  Extra digit of precision this way, 2237 / 10 or 223.7.
 *  $ units
 *  You have: 0.1 meter / millisecond
 *  You want: mile / hour
 *   * 223.69363
 */
static void
weather_station()
{
	word msec, n, dir, mph;

	msec = anemometer(&n);
	if (!msec)
		return;                 /* eek? too soon (or too long since) */

	if (n) {
		mph = muldiv(n, cfg.anemoscale, msec) / 10;
		have_anemometer = TRUE;
	} else
		mph = 0;

	dir = (vane_to_deg[windvane()] + cfg.vaneoffset) % 360u;

	log_wind_sample(dir, n, mph);

	if (cfg.wx_dump) {
		print_string("wx  ");

		print_short(dir);              /* not the averaged direction */
		print_string(STR_DEGREE "  ");
		print_short(miles_to_km(mph));
		print_string(" km/h");

		if (have_thermometer) {
			print_string("  ");
			print_signed_short(temperature_celsius_lo);
			if (temperature_celsius_lo != temperature_celsius_hi) {
				print_string("/");
				print_signed_short(temperature_celsius_hi);
			}
			print_string(STR_DEGREE "C");
		}
		if (have_hygrometer) {
			print_string("  ");
			print_short(relative_humidity);
			print_string("%");
		}
		if (have_barometer) {
			print_string("  ");
			print_short(barometric_pressure);
			print_string("hPa");
		}
		eol();
	}
}

/*
 *  HIH-3610
 *
 *  sensor RH = (Vad/Vdd - 0.16) / 0.0062;
 *  temp compensation = 1.0546 - 0.00216 * celsius;
 *
 *  assume Vdd is 10000h and Vad (AN7) is thus already the ratio.
 *  typical Vad 0.8v ... 3.9v is 41 ... 199 decimal (MSByte).
 *
 *  'rh'   is unsigned 0...FFC0h from HIH-3610 sensor
 *  'temp' is signed celsius from 1st DS1820 (or something)
 */
static word
hih3610_relhum(word rh, short temp)
{
	if (temp <  -50)       /* huh ? */
		temp =  -50;
	else
	if (temp > +100)
		temp = +100;

	if (rh >= 10486)
		rh -= 10486;
	else
		rh = 0;

	rh = muldiv(rh, 16129, 34557 - 70 * temp);

	rh = (rh + 100) / 200;      /* round and remove 100/0.5 scale */

	return (rh > 100 ? 100 : rh);
}

static void
hih3610_round()
{
	/* temperature needed too */

	if (have_thermometer) {
		word ana = read_AN(7);

		/* sensor likely missing, if A/D is 0 (AN7 grounded by default) */

		if (ana)
			hygrometer(hih3610_relhum(ana, temperature_celsius_lo));
	}
}

#define LM75 0x90 /* or LM76 or ... */

static void
lm75_round()
{
	short half_c;

	if (i2c_rd(LM75, &half_c, sizeof(half_c)))
		thermometer(half_c);
}

#define DS1621 0x9E /* or ds1721/1631/1731/... */

void
ds1621_round()
{
	static byte
		Access_Config    = 0xAC,
		Read_Temperature = 0xAA,
		Start_Convert_T  = 0xEE, /* or 0x51, see below */
		happy = FALSE;

	short temp;
	byte config;

	if (happy
	 && i2c_wr(DS1621 | NOSTOP, &Read_Temperature, 1)
	 && i2c_rd(DS1621, &temp, sizeof(temp))) {
		thermometer(temp >> 7);                /* half_c */
		return;
	}

	/* start or restart */

	happy = FALSE;

	if (i2c_wr(DS1621 | NOSTOP, &Access_Config, 1) == FALSE
	 || i2c_rd(DS1621, &config, 1) == FALSE)
		return;

	if ((config & 0x0C) == 0x0C)
		Start_Convert_T = 0x51;    /* later family */

	if (config & 0x01) {
		config &= ~0x01;           /* continuous, not 1SHOT */

		if (i2c_wr(DS1621 | NOSTOP, &Access_Config, 1) == FALSE
		 || i2c_wr(NOSTART, &config, 1) == FALSE)
			return;
	}
	if (i2c_wr(DS1621, &Start_Convert_T, 1))
		happy = TRUE;
}

/* APRS weather report */

static word
winddir_cse()
{
	word dir;

	if ((dir = winddir()) == 0)
		dir = 360;

	return (dir);
}

static word
windspeed_spd()
{
	return (windspeed()); /* XXX mph ? spec says knots, but ... */
}

byte *
encode_weather_ascii(byte *p)
{
	/*
	 * fixed part:
	 *
	 * 001...360 wind direction, degrees
	 * /999      speed, 1 minute sustained, mph (spec says knots ?)
	 * g999      gust peak, 5 minute, mph
	 * t999      temperature, fahrenheit. t-99 if negative XXX outdoor temp
	 */
	if (have_anemometer) {
		p = sprint_03u(p, winddir_cse());
		*p++ = '/';
		p = sprint_03u(p, windspeed_spd());
		*p++ = 'g';
		p = sprint_03u(p, windgust());
	} else
		p = strcpyx(p, ".../...g...");

	*p++ = 't';
	if (have_thermometer)
		sprint_03d_signed(p, temperature_fahrenheit_lo); /* outdoor temp */
	else
		strcpyx(p, "...");
	p += 3;

	/*
	 * variable part:
	 *
	 * relative humidity, percent - h99, h00 if 100%
	 * barometric pressure, .1 millibars - b99999, that is 5 digits
	 * rainfall in 1 hour, 1/100 inch - r000
	 * rainfall since midnight - P000
	 * water level, centimeter - F[+-]999, that is -999 ... +000 ... +999 (?)
	 * fuel temperature F
	 * fuel moisture f
	 * voltage, deciVolt - V999 for 99.9 volt
	 */
	if (have_hygrometer) {
		p = sprint_03u(p, relative_humidity);
		p[-3] = 'h';
	}
	if (have_barometer) {
		*p++ = 'b';
		p = sprint_u(p, barometric_pressure);
		*p++ = '0';
	}
	if (have_rain_gauge) {
		*p++ = 'r';
		p = sprint_03u(p, rain_gauge_centi_inch(rainfall_1hr));
	}

	return (p);
}

/*
 *  after every WX_INTERVAL seconds, call sensors and such.
 *  calls are dispersed, one per weather_run().
 */
void
weather_run()
{
	extern void ow_round(), sht1x_round();

	static void (*match[])() = {
		ow_round,
		sht1x_round,
		lm75_round,
		ds1621_round,
		hih3610_round,    /* after any thermometers (tcomp) */

		weather_station,  /* must be last */
		NULL
	}, (**game)() = match;

	static stamp mark;

	if (*game)
		fncall(0, **game++);

	else if (time_past(mark)) {
		game = match;
		thermometer_hilo_rewind();
		future_msec(&mark, WX_INTERVAL * 1000);
	}
}

/*
 *  if sensor exists, it disappears after some time.
 *  initialize have_xxx to 256 - 30 or something like that.
 */
void
sensors_timeout()
{
	if (have_anemometer)
		have_anemometer++;

	if (have_barometer)
		have_barometer++;

	if (have_hygrometer)
		have_hygrometer++;

	if (have_thermometer)
		have_thermometer++;
}

void
weather_midnight()
{
	rainfall_today = 0;
}

